##  hmsc.tcl
##
##		This file implements code for
##		canvas operations
##
##  Copyright (c) 1998 <Irrational Numbers> : <buraldo@maganet.net> 
##
##  This program is free software; you can redistibute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation.
##
##  See the file "license" for information on usage and
##  redistribution of this program, and for a disclaimer of all warranties.
##
##  VER 1.1



	##
	## global little procs
	##

proc hasTag {w id tag} {
	return [expr [lsearch -exact [$w gettags $id] $tag] >=0]
}

proc lgrep { L pattern } {
	set i [lsearch -glob $L $pattern]
	return  [lindex $L $i]
}

proc lremove { listName elem } {
	upvar $listName L

	set i [lsearch -exact $L $elem]
	set L [lreplace $L $i $i]
	return
}

proc lunique {L} {
	set L [lsort $L]
	set R {}
	set prev [lindex $L 0]
	lappend R $prev
	foreach el $L {
		if {$el != $prev} { lappend R $el }
		set prev $el
	}
	return $R
}

proc min { a b } { return [expr ($a<$b) ? $a : $b] }
proc max { a b } { return [expr ($a>$b) ? $a : $b] }
 


	##
	## SerialNo namespace
	##

namespace eval SerialNo {

 variable SerialNo 0
 proc getNext {} {
	variable SerialNo
	incr SerialNo
	return $SerialNo
 }
} ; ## = end of namespace =====================================================



	##
	## CVS namespace
	##

namespace eval CVS {

variable ui_msgline 

	# update whenever scrollregion change !
variable XMIN
variable XMAX
variable YMIN
variable YMAX

proc get_paper {w} {
	variable XMIN
	variable XMAX
	variable YMIN
	variable YMAX
	return [list $XMIN $YMIN $XMAX $YMAX]
}

proc set_paper {w x0 y0 x1 y1} {
	variable XMIN
	variable XMAX
	variable YMIN
	variable YMAX
	set XMIN $x0 ; set YMIN $y0 ; set XMAX $x1 ; set YMAX $y1
	$w configure -scrollregion \
		[list [expr $XMIN-5] [expr $YMIN-5] [expr $XMAX+5] [expr $YMAX+5]] 
	global pref
	set pref(DX) [expr $XMAX - $XMIN]
	set pref(DY) [expr $YMAX - $YMIN]
	draw_paper $w
}

proc draw_paper {w} {
	variable XMIN
	variable XMAX
	variable YMIN
	variable YMAX
	$w delete "paperFrame"	
	$w create rectangle [expr $XMIN-2] [expr $YMIN-2] \
		[expr $XMAX+2] [expr $YMAX+2] -outline gray -tags "paperFrame"
}


proc set_DocTitle {w txt} {
	set id [$w create text 1 1 -text "MSC " -anchor nw ]
	set xyxy [$w bbox $id]
	set x [lindex $xyxy 2]
	set y [lindex $xyxy 3]
	
	set id [$w create text $x $y -text $txt \
		-anchor sw \
		-tags "editable docTitle"]
}

proc get_DocTitle {w} {
	return [$w itemcget "docTitle" -text]
}

proc node {w kind x y {dx 0} {dy 0} {txt ""} } {
	set grp [create_sym_node  $w $kind $x $y $dx $dy $txt]

	variable $grp.arc
	set $grp.arc(out) {}
	set $grp.arc(in)  {}
	return $grp
}

proc node_delete { w nodeGrp_List } {
	foreach nodeGrp $nodeGrp_List {
		variable $nodeGrp.arc
		if { [hasTag $w "node&&$nodeGrp" startSymb] } continue
		foreach a [concat [set $nodeGrp.arc(out)] [set $nodeGrp.arc(in)] ] {
			arc_delete $w $a
		}
		unset $nodeGrp.arc
		$w delete $nodeGrp
	}
}

proc redraw_sym_node { w nodeGrp DX DY } {
    set boxID [$w find withtag "node&&$nodeGrp"]
	set args [get_node $w $nodeGrp]	
	set kind [lindex $args 0]
	set xc   [lindex $args 1]
	set yc   [lindex $args 2]
	switch $kind {
	 startSymb
		{
			# triangle turned upside down
		$w coords $boxID $xc [expr $yc+$DY/2] 	\
				[expr $xc-$DX/2] [expr $yc-$DY/2] 	\
				[expr $xc+$DX/2] [expr $yc-$DY/2]
		}
	 endSymb
		{
			# triangle
		$w coords $boxID $xc [expr $yc-$DY/2] 	\
				[expr $xc-$DX/2] [expr $yc+$DY/2] 	\
				[expr $xc+$DX/2] [expr $yc+$DY/2]
		}
	 connectSymb
		{
			# oval
		$w coords $boxID [expr $xc-$DX/2] [expr $yc-$DY/2] \
				[expr $xc+$DX/2] [expr $yc+$DY/2]
		}
	 mscrefSymb
		{
			# rectangle 
		$w coords $boxID [expr $xc-$DX/2] [expr $yc-$DY/2]\
			 	[expr $xc+$DX/2] [expr $yc+$DY/2]
		}
	 conditionSymb
		{
			# rectangle with cutten sides (?)
		set d [min [expr $DY/4] [expr $DX/2]]
		set xA [expr $xc -$DX/2]
		set xB [expr $xc -($DX/2-$d)]
		set xC [expr $xc +($DX/2-$d)]
		set xD [expr $xc +$DX/2]
		set yA [expr $yc -$DY/2]
		set yB [expr $yc +$DY/2]
		$w coords $boxID  $xA $yc  $xB $yA  $xC $yA  $xD $yc \
				$xC $yB  $xB $yB 
		}
	 default
		{
		}
	}
}


proc create_sym_node { w kind x y DX DY txt } {
	set grp grp#[SerialNo::getNext]
	switch $kind {
	 startSymb		-
	 endSymb		{
		set ID [$w create polygon $x $y $x $y $x $y -fill gray75 -outline black]
		if {$DX == 0} {set DX 20}
		if {$DY == 0} {set DY 20}
		set resizeable 0
		}
	 connectSymb	{
		set ID [$w create oval $x $y $x $y -fill gray75]
		if {$DX == 0} {set DX 10}
		if {$DY == 0} {set DY 10}
		set resizeable 0
		}
	 mscrefSymb		{
		set ID [$w create rectangle $x $y $x $y -fill gray75]
		if {$DX == 0} {set DX 60}
		if {$DY == 0} {set DY 30}
		set resizeable 1
		if { $txt == "" } { set txt "*mscref*" }
		$w create text $x $y -text $txt -tags "editable $grp text&&$grp"
		}
	 conditionSymb	{
		set ID [$w create polygon $x $y $x $y $x $y $x $y $x $y $x $y \
			-fill gray75 -outline black]
		if {$DX == 0} {set DX 60}
		if {$DY == 0} {set DY 30}
		set resizeable 1
		if { $txt == "" } { set txt "*condition*" }
		$w create text $x $y -text $txt -tags "editable $grp text&&$grp"
		}
	 default		{
		tk_messageBox -title "Internal Error" -message "$kind: unknown symbol"
		return ""
		}
	}
	  ## remember: the order of tags is important for bindings eval order!
	$w addtag "node" withtag $ID
	$w addtag $kind withtag $ID
	$w addtag $grp withtag $ID
	$w addtag "node&&$grp" withtag $ID
	if {$resizeable} {
		$w addtag "resizeableNode" withtag $ID
	}
	# -----------------------------
	redraw_sym_node $w $grp $DX $DY
	# -----------------------------
	return $grp
}

proc get_kind {w grp} {
	return [lgrep [$w gettags "node&&$grp"] "*Symb"]
}

proc get_node {w grp} {
	set kind [get_kind $w $grp]
	  ## PATCH: bbox is not suitable for polyline ...
	  ## nor for rectangle ..
	  ## nor for triangle ..
	set coord [$w coords "node&&$grp"]
	if { $kind == "conditionSymb" } {
		set x0 [lindex $coord 0]
		set y0 [lindex $coord 3]
		set x1 [lindex $coord 6]
		set y1 [lindex $coord 9]
	} elseif { $kind == "mscrefSymb" } {
		set x0 [lindex $coord 0]
		set y0 [lindex $coord 1]
		set x1 [lindex $coord 2]
		set y1 [lindex $coord 3]
	} elseif { $kind == "startSymb" } {
		set x0 [lindex $coord 2]
		set y0 [lindex $coord 3]
		set x1 [lindex $coord 4]
		set y1 [lindex $coord 1]
	} elseif { $kind == "endSymb" } {
		set x0 [lindex $coord 2]
		set y0 [lindex $coord 1]
		set x1 [lindex $coord 4]
		set y1 [lindex $coord 3]
	} else {
		set xyxy [$w bbox "node&&$grp"]
		set x0 [lindex $xyxy 0]
		set y0 [lindex $xyxy 1]
		set x1 [lindex $xyxy 2]
		set y1 [lindex $xyxy 3]
	}
	 # always the midpoint
	set xc [expr ($x0+$x1)/2]
	set yc [expr ($y0+$y1)/2]
	set dx [expr $x1-$x0]
	set dy [expr $y1-$y0]
	set txt [$w itemcget "text&&$grp" -text]

	return [concat $kind $xc $yc $dx $dy \{ $txt \}]
}

proc new_arc {w N1} {
	set N1params [get_node $w $N1]
	 set x0 [lindex $N1params 1]
	 set y0 [lindex $N1params 2]
	variable ui_msgline
	set ui_msgline "click on a node  or  press <Esc> to stop"

	set rubberID [$w create line $x0 $y0 $MUI::click(X) $MUI::click(Y) \
		 -fill orange -arrow last]
	$w lower $rubberID
        MUI::enable_rubberLine $w $rubberID
	$w delete $rubberID
	set ui_msgline ""

	if { $MUI::click(abandon) } return
	set N2 $MUI::click(node_picked)
	return [arc $w $N1 $N2]
}

proc create_sym_arc { w } {
	set grp grp#[SerialNo::getNext]
	 # draw a dummy line; it will be redrawn later ...
	set ID [$w create line 0 0 0 0 -arrow last]
	$w addtag "arc" withtag $ID
	$w addtag $grp withtag $ID
	return $grp
}


proc arc {w node1 node2 {side left}} {
	set arcGrp [create_sym_arc $w]

	variable $node1.arc
	variable $node2.arc
	lappend $node1.arc(out) $arcGrp
	lappend $node2.arc(in)  $arcGrp

	variable $arcGrp.node
	variable $arcGrp.curve
	set $arcGrp.node(out) $node1
	set $arcGrp.node(in)  $node2
	set $arcGrp.curve(flag) 0 ; # don't care now
	set $arcGrp.curve(side) $side
	arc_redraw $w $arcGrp
	return $arcGrp
}

proc get_arc { w grp } {
	variable $grp.node
	variable $grp.curve
	return [list [set $grp.node(out)] [set $grp.node(in)] [set $grp.curve(side)] ]
}

proc arc_is_curve { arcGrp } {
	variable $arcGrp.curve
	return [set $arcGrp.curve(flag)]
}

proc arc_flip { w arcGrp } {
	puts "grp : $arcGrp"
	variable $arcGrp.curve
	if { [set $arcGrp.curve(side)] == "left" } {
		set $arcGrp.curve(side) right
	} else {
		set $arcGrp.curve(side) left
	}
	arc_redraw $w $arcGrp 
}

variable arcOffset 10
variable arcSmooth 1

proc arc_redraw { w arcGrp } {
	variable $arcGrp.node
	variable $arcGrp.curve
	variable arcOffset
	variable arcSmooth

	set nodeA [set $arcGrp.node(out)]
	set nodeB [set $arcGrp.node(in)]

	set xyxy [$w bbox $nodeA]
	set A_x1 [lindex $xyxy 0]
	set A_x2 [lindex $xyxy 2]
	set A_y1 [lindex $xyxy 1]
	set A_y2 [lindex $xyxy 3]
	## outgoing point is always below !
	set A_xM [expr ($A_x1 + $A_x2) /2]
	set A_yM $A_y2

	set xyxy [$w bbox $nodeB]
	set B_x1 [lindex $xyxy 0]
	set B_x2 [lindex $xyxy 2]
	set B_y1 [lindex $xyxy 1]
	set B_y2 [lindex $xyxy 3]
	## ingoing point is always above !
	set B_xM [expr ($B_x1 + $B_x2) /2]
	set B_yM $B_y1

	if { $B_yM > $A_yM } {
		set $arcGrp.curve(flag) 0
		$w coords $arcGrp $A_xM $A_yM $B_xM $B_yM
		$w itemconfigure $arcGrp -smooth 0 -arrow none
	} else {
		set $arcGrp.curve(flag) 1
		set side [set $arcGrp.curve(side)]
		if { 
			$side == "left"  &&  $A_xM < $B_xM 
			||
			$side == "right"  &&  $A_xM >= $B_xM 
		 } {
			set yT1 [expr $A_yM +$arcOffset]
			set yT2 [expr [min $A_y1 $B_yM] -$arcOffset]
		} else {
			set yT1 [expr [max $A_yM $B_y2] +$arcOffset]
			set yT2 [expr $B_yM -$arcOffset]

		}
		if { $side == "left" } {
			set xT [expr [min $A_x1 $B_x1] -$arcOffset]
		} else {
			set xT [expr [max $A_x2 $B_x2] +$arcOffset]
		}
		$w coords $arcGrp $A_xM $A_yM  $A_xM $yT1 \
			$xT $yT1  $xT $yT2 \
			$B_xM $yT2  $B_xM $B_yM
		$w itemconfigure $arcGrp -smooth $arcSmooth -arrow last
	}
}

proc arc_delete { w arcGrp } {
	 ## this test is for arcs joining the same node;
	 ##		you cannot delete them two times !
	if { ! [info exist CVS::$arcGrp.node] } return

	variable $arcGrp.node
	variable $arcGrp.curve
	set n1 [set $arcGrp.node(out)]
	variable $n1.arc
	lremove $n1.arc(out) $arcGrp
	set n2 [set $arcGrp.node(in)]
	variable $n2.arc
	lremove $n2.arc(in)  $arcGrp
	unset $arcGrp.node
	unset $arcGrp.curve

	$w delete $arcGrp
}


proc node_move {w nodeGrp dx dy} {
	$w move $nodeGrp $dx $dy		
	variable $nodeGrp.arc
	foreach a [set $nodeGrp.arc(out)] {
		arc_redraw $w $a 
	}
	foreach a [set $nodeGrp.arc(in)] {
		arc_redraw $w $a
	}
}

# ============================================================================
#	ASCII Output (Z.120)
# ============================================================================

  ## transform string from "grp#..." to "L#..."
proc node2label {str} {
	return L#[string range $str 4 end]
}


  ## transform each node identifier (grp#...) in a pretty look (label#...)
  ## and separate each label with the keyword "alt"
proc labelList { w node } {
	variable $node.arc
	set L [set $node.arc(out)]
	set first [lindex $L 0]
	if { $first == {} } { return {} }
	set tail  [lrange $L 1 end]
	variable $first.node
	set R [node2label [set $first.node(in)]]
	foreach a $tail {
		variable $a.node
		lappend R "ALT"
		lappend R [node2label [set $a.node(in)]]
	}
	return $R
}


proc write_nodeExpr { fID cvs node } {
	global marked

	set errCode 0
	if { [info exist marked($node)] } {return $errCode}
	set marked($node) 1

	puts -nonewline $fID "  [node2label $node] : "

	set kind [get_kind $cvs $node]
	  ## remove last 4 chars of $kind (i.e. remove "Symb")
	set kindStr [string range $kind 0 [expr [string length $kind]-1 -4]]
	if { $kindStr == "end" } {
		 puts $fID "end ;"
		 return $errCode
	}

		## nota: ne deriva che un MSC non si puo'chiamare "connect" "condition" "empty"

	## check: if succ is empty, it's an error
	variable $node.arc
	if { [set $node.arc(out)] == {} } {
		tk_messageBox -title "Graph Error" \
			-message "Found a node without successors\nOperation canceled"
		set errCode 1
		return $errCode
	}
	switch $kindStr {
	 condition
		{
		set txt [$cvs itemcget "text&&$node" -text]
		set txt "condition $txt"
		}
	mscref
		{
		set txt [$cvs itemcget "text&&$node" -text]
		if { [llength $txt] > 1 } {
			set txt "( $txt )"
		} 
		}
	connect
		{
		set txt "connect"
		}
	}
	puts $fID "$txt  seq( [labelList $cvs $node] ) ;"

	variable $node.arc
	foreach a [set $node.arc(out)] {
		variable $a.node
		set n [set $a.node(in)]
	
		set errCode [write_nodeExpr $fID $cvs $n]
		if {$errCode} break
	}
	return $errCode
}

	## to do: check nodi appesi...
proc writeZ120 { fID cvs } {
	global HMSC_VER
	global marked

	set errCode 0

	puts $fID "/* HMSC version $HMSC_VER */"
	puts $fID "msc [get_DocTitle $cvs] ;"
	# get start node:
	set nodeGrp [lgrep [$cvs gettags "startSymb"] grp#*]
	puts $fID "expr [labelList $cvs $nodeGrp] ;"
	catch {unset marked}  ; #clean is better
	set marked($nodeGrp) 1

	variable $nodeGrp.arc
	foreach a [set $nodeGrp.arc(out)] {
		variable $a.node
		set n [set $a.node(in)]
	
		set errCode [write_nodeExpr $fID $cvs $n]
		if { $errCode } break
	}
	puts $fID "endmsc ;"
	unset marked
	return $errCode
}
	# =======================================================================

	#
	# Graph checking
	#
proc check_node { cvs node } {
	global marked

	set errCode 0
	if { [info exist marked($node)] } {return 0}
	set marked($node) 1

	set kind [get_kind $cvs $node]
	if { $kind == "endSymb" } {
		 return 0
	}

	## check: if succ is empty, it's an error
	variable $node.arc
	if { [set $node.arc(out)] == {} } {
		$cvs itemconfigure "node&&$node" -fill blue
		return 1
	}

	variable $node.arc
	foreach a [set $node.arc(out)] {
		variable $a.node
		set n [set $a.node(in)]
	
		set status [check_node $cvs $n]
		if {$status} {set errCode 1}
	}
	return $errCode
}

proc checkDiagram { cvs } {
	global marked

	variable ui_msgline
	set ui_msgline "Graph Checking ..."
	set errCode 0
	# get start node:
	set nodeGrp [lgrep [$cvs gettags "startSymb"] grp#*]
	catch {unset marked}  ; #clean is better
	set marked($nodeGrp) 1

	variable $nodeGrp.arc
	foreach a [set $nodeGrp.arc(out)] {
		variable $a.node
		set n [set $a.node(in)]
	
		set status [check_node $cvs $n]
		if { $status } { set errCode 1 }
	}
	# finally, check for unconnected nodes ..
	foreach id [$cvs find withtag "node"] { 
		set nodeGrp [lgrep [$cvs gettags $id] grp#*]
		if { ! [info exist marked($nodeGrp)] } {
			set errCode 1
			$cvs itemconfigure $id -fill blue
		}
	}
	set ui_msgline ""
	if { $errCode } {
		tk_messageBox -title "Graph Checking" \
			-message "Found errors: look at the blue nodes.."
	} else {
		tk_messageBox -title "Graph Checking" \
			-message "pretty good"
	}
	unset marked
	return $errCode
}

} ; ## = end of namespace =====================================================
